"""A sample Configurable-like module.

-----

This software is released under an MIT/X11 open source license.

Copyright 2014 Diffeo, Inc.

"""

config_name = 'configurable'
default_config = { 'type': 'module' }
